#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096
#define MAXSTATUS 1024
#define MAXTILES 4096

#ifdef __GNUC__
#define BPACK __attribute__ ((packed))
#else
#define BPACK
#endif

#ifdef _MSC_VER
#pragma pack(1)
#endif

#ifdef __WATCOMC__
#pragma pack(push,1);
#endif

typedef struct BPACK
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingpanning;
	signed char ceilingshade;
	int ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorpanning;
	signed char floorshade;
	int floorz;
	short floorpicnum, floorheinum;
	int tag;
} sectortype;

typedef struct BPACK
{
	int x, y;
	short point2;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat, panning;
	short picnum, overpicnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	int tag;
} walltype;

typedef struct BPACK
{
	int x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	int tag;
	char *extra;
} spritetype;

typedef struct BPACK
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingxpanning, ceilingypanning;
	signed char ceilingshade;
	int ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorxpanning, floorypanning;
	signed char floorshade;
	int floorz;
	short floorpicnum, floorheinum;
	int tag;
} newsectortype;

typedef struct BPACK
{
	int x, y;
	short point2;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat, xpanning, ypanning;
	short picnum, overpicnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	int tag;
} newwalltype;

typedef struct BPACK
{
	int x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, xvel, yvel, zvel, owner;
	short sectnum, statnum;
	int tag;
	char *extra;
} newspritetype;

#ifdef _MSC_VER
#pragma pack()
#endif

#ifdef __WATCOMC__
#pragma pack(pop)
#endif

//**************************************************************************

static sectortype sector[MAXSECTORS];
static walltype wall[MAXWALLS];
static spritetype sprite[MAXSPRITES];

static newsectortype newsector[MAXSECTORS];
static newwalltype newwall[MAXWALLS];
static newspritetype newsprite[MAXSPRITES];

static int posx, posy, posz, mapversion;
static short int ang, cursectnum, numsectors, numwalls, numsprites;

int loadoldboard (char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&mapversion,4);
	if (mapversion != 3L) return(-2);
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&cursectnum,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(spritetype));
	close(fil);

	return(0);
}

int savenewboard (char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

	write(fil,&mapversion,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) write(fil,&newsector[i],sizeof(newsectortype));
	for(i=0;i<numwalls;i++) write(fil,&newwall[i],sizeof(newwalltype));
	for(i=0;i<numsprites;i++) write(fil,&newsprite[i],sizeof(newspritetype));
	close(fil);

	return(0);
}

int main (int argc, char **argv)
{
	char ch, filename[80];
	int i, j;

	if (argc != 2)
	{
		printf("CONVMAP4 [filename]\n");
		printf("This CONVMAP4.EXE converts map version 3 to map version 4\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	printf("Converting from map version 3 to map version 4...\n");
	printf("Adding new fields to structures...\n");

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("Cannot convert %s.\n",filename);
		exit(0);
	}

	for(i=0;i<numsectors;i++)
	{
		newsector[i].wallptr = sector[i].wallptr;
		newsector[i].wallnum = sector[i].wallnum;
		newsector[i].ceilingstat = sector[i].ceilingstat & ~4;
		if ((sector[i].ceilingstat&4) == 0)
		{
			newsector[i].ceilingxpanning = sector[i].ceilingpanning;
			newsector[i].ceilingypanning = 0;
		}
		else
		{
			newsector[i].ceilingxpanning = 0;
			newsector[i].ceilingypanning = sector[i].ceilingpanning;
		}
		newsector[i].ceilingshade = sector[i].ceilingshade;
		newsector[i].ceilingz = sector[i].ceilingz;
		newsector[i].ceilingpicnum = sector[i].ceilingpicnum;
		newsector[i].ceilingheinum = sector[i].ceilingheinum;
		newsector[i].floorstat = sector[i].floorstat & ~4;
		if ((sector[i].floorstat&4) == 0)
		{
			newsector[i].floorxpanning = sector[i].floorpanning;
			newsector[i].floorypanning = 0;
		}
		else
		{
			newsector[i].floorxpanning = 0;
			newsector[i].floorypanning = sector[i].floorpanning;
		}
		newsector[i].floorshade = sector[i].floorshade;
		newsector[i].floorz = sector[i].floorz;
		newsector[i].floorpicnum = sector[i].floorpicnum;
		newsector[i].floorheinum = sector[i].floorheinum;
		newsector[i].tag = sector[i].tag;
	}
	for(i=0;i<numwalls;i++)
	{
		newwall[i].x = wall[i].x;
		newwall[i].y = wall[i].y;
		newwall[i].point2 = wall[i].point2;
		newwall[i].cstat = wall[i].cstat & ~64;
		newwall[i].shade = wall[i].shade;
		newwall[i].xrepeat = wall[i].xrepeat;
		newwall[i].yrepeat = wall[i].yrepeat;
		if ((wall[i].cstat&64) == 0)
		{
			newwall[i].xpanning = wall[i].panning;
			newwall[i].ypanning = 0;
		}
		else
		{
			newwall[i].xpanning = 0;
			newwall[i].ypanning = wall[i].panning;
		}
		newwall[i].picnum = wall[i].picnum;
		newwall[i].overpicnum = wall[i].overpicnum;
		newwall[i].nextsector1 = wall[i].nextsector1;
		newwall[i].nextwall1 = wall[i].nextwall1;
		newwall[i].nextsector2 = wall[i].nextsector2;
		newwall[i].nextwall2 = wall[i].nextwall2;
		newwall[i].tag = wall[i].tag;
	}
	for(i=0;i<numsprites;i++)
	{
		newsprite[i].x = sprite[i].x;
		newsprite[i].y = sprite[i].y;
		newsprite[i].z = sprite[i].z;
		newsprite[i].cstat = sprite[i].cstat;
		newsprite[i].shade = sprite[i].shade;
		newsprite[i].xrepeat = sprite[i].xrepeat;
		newsprite[i].yrepeat = sprite[i].yrepeat;
		newsprite[i].picnum = sprite[i].picnum;
		newsprite[i].ang = sprite[i].ang;
		newsprite[i].xvel = 0;
		newsprite[i].yvel = 0;
		newsprite[i].zvel = sprite[i].zvel;
		newsprite[i].owner = sprite[i].owner;
		newsprite[i].sectnum = sprite[i].sectnum;
		newsprite[i].statnum = sprite[i].statnum;
		newsprite[i].tag = sprite[i].tag;
		newsprite[i].extra = sprite[i].extra;
	}

	for(i=0;i<numwalls;i++)
		if (newwall[i].nextwall1 >= 0)
		{
			if ((newwall[i].cstat&16) > 0)
			{
				if (newwall[i].overpicnum == 0)
				{
					newwall[i].overpicnum = newwall[newwall[i].nextwall1].picnum;
				}
			}
		}

	mapversion = 4L;
	if (savenewboard(filename) == 0)
		printf("%s successfully converted.\n",filename);
	else
		printf("%s NOT converted for some stupid reason.\n",filename);
	return(0);
}
